#include <hmg.ch>

#include "hbclass.ch"

Class GraphPlus

   Data aData          INIT {}    // Graph data
   Data aLegends       INIT {}
   Data aCategories    INIT {}  // Horizontal Axis
   Data nGraphType     INIT 1  // COLUMN
   Data cTitle         INIT '' // Graph Title
   Data cXTitle        INIT '' // Horizontal Title
   Data cYTitle        INIT '' // Vertical Title
   Data lXRotate       INIT .f. // xValues Rotate
   Data nHeight        INIT 600 // Graph Height
   Data nWidth         INIT 600 // Graph Width
   Data lShowLegends   INIT .f. // Show legends
   Data lDataLabels    INIT .f. // show data labels
   Data lHGrid         INIT .t. // Horizontal Grid
   Data lVGrid         INIT .t. // Vertical Grid
   Data lShowValues    INIT .t. // Show Value Axis
   Data lShowCategories INIT .t. // show categories
   Data nLegendPos     INIT 1   // 0 Right 1 Bottom
   Data lOutBorder     INIT .t. // draw outside border
   Data aColors        INIT {}  // aColors
   Data aTitleFont     INIT { 'Arial', 12, .t., .f., .f., .f. }
   Data aLegendFont    INIT { 'Arial', 8, .f., .f., .f., .f. }
   Data aBackColor     INIT { 255, 255, 255 }
   Data aTitleColor    INIT { 0, 0, 255 }
   Data aTextColor      INIT { 0, 0, 0 }
   Data aGridLineColor INIT { 200, 200, 200 }
   Data nLegendBoxSize INIT 10
   Data nLegendWidth   INIT 80
   Data nColorTheme    INIT 1
   Data nUpper         INIT 0
   Data nLower         INIT 0
   Data lCategoryBelow INIT .t.
   Data nRange         INIT 0
   Data nPixelperPoint INIT 0
   Data nLineWidth     INIT 2
   Data nZeroPos       INIT 0
   Data nVDivision     INIT 5
   Data nHDivision     INIT 0
   Data nUpperDivision INIT 0
   Data nLowerDivision INIT 0
   Data cPicture       INIT '99999'
   Data nCategoryRow   INIT 0
   Data nBarGapRatio   INIT 0.2 // in percentage 20%
   Data nGapWidthRatio INIT 1  // in percentage 200%
   Data nImageHeight   INIT 600
   Data nImageWidth    INIT 600
   Data nPointSize     INIT 10
   
   Data hBitMap        INIT nil // Bitmap to store Graph
   Data hDC            INIT nil 
   Data BTStruct       INIT nil  
   Data nTop           INIT 10
   Data nLeft          INIT 10
   Data nRight         INIT 590 
   Data nBottom        INIT 590
   Data aScale         INIT {}
   Data aColorThemes   INIT { { { 68, 115, 197 }, { 237, 125, 49 }, { 165, 165, 165 }, { 253, 193, 1 }, { 90, 155, 213 }, { 112, 173, 70 } },;
                              { { 68, 115, 197 }, { 165, 165, 165 }, { 90, 155, 213 }, {38, 68, 120 }, { 99, 99, 99 }, { 37, 95, 145 } },;
                              { { 237, 125, 49 }, { 253, 193, 1 }, { 112, 173, 70 }, { 160, 71, 13 }, { 152, 115, 0 }, { 66, 104, 43 } },;
                              { { 112, 173, 70 }, { 90, 155, 213 }, { 253, 193, 1 }, { 66, 104, 43 }, { 37, 95, 145 }, { 152, 115, 0 } },;
                              { { 68, 115, 197 } },;
                              { { 237, 125, 49 } },;
                              }
   Data lLighter       INIT .t.
   
   Method New 
   Method Draw
   Method Bitmap
   Method DrawColumnGraph
   Method DrawBarGraph
   Method DrawLineGraph
   Method DrawPieGraph
   Method DrawPointsGraph
   Method DrawLegends
   Method InitColors
   Method DrawXYTitles
   Method DrawCategories
   Method DrawVerticalScale
   Method DrawHorizontalGrid
   Method DrawVerticalGrid
   Method SwitchRowsToColumns
   Method Save
   Method GraphData    SETGET
   Method GridData     SETGET
   Method Legends      SETGET
   Method Picture      SETGET
   Method Categories   SETGET
   Method GraphColors  SETGET
   Method GraphType    SETGET
   Method Title        SETGET
   Method XTitle       SETGET
   Method YTitle       SETGET
   Method Height       SETGET
   Method Width        SETGET
   Method ShowLegends  SETGET
   Method ShowLabels   SETGET
   Method ShowHGrid    SETGET
   Method ShowVGrid    SETGET
   Method ShowValues   SETGET
   Method LegendPos    SETGET
   Method ColorTheme   SETGET
   Method BackColor    SETGET

End class   
   
METHOD New CLASS GraphPlus
RETURN self

METHOD GraphData( aData ) CLASS GraphPlus
   local i := 0, nLen := 0
   if PCount() == 0
      return ::aData
   else
      if valtype( aData[ 1 ] ) == 'A' // 2D array
         ::aData := aData
      else
         ::aData := { aData } // 1D - make it 2D
      endif      
   endif
   return Self
   
Method GridData( cFormName, cControlName ) Class GraphPlus
   local i := 0
   local j := 0
   local nRowCount := getproperty( cFormName, cControlName, 'ITEMCOUNT' )
   local nColumnCount := 0
   local aLineData := {}
   local aLine := {}
   if PCount() == 0
      return ::aData
   else
      asize( ::aData, 0 )
      asize( ::aLegends, 0 )
      asize( ::aCategories, 0 )
      nColumnCount := getproperty( cFormName, cControlName, 'COLUMNCOUNT' )
      for i := 1 to nColumnCount
         if i == 1
            ::cYTitle := getproperty( cFormName, cControlName, 'COLUMNHEADER', i )
         else   
            aadd( ::aCategories, getproperty( cFormName, cControlName, 'COLUMNHEADER', i ) )
         endif   
      next i
      if nRowCount > 0
         for i := 1 to nRowCount
            aLineData := getproperty( cFormName, cControlName, 'ITEM', i )
            asize( aLine, 0 )
            for j := 1 to nColumnCount
               if j == 1
                  aadd( ::aLegends, aLineData[ j ] )
               else   
                  aadd( aLine, val( alltrim( aLineData[ j ] ) ) )
               endif
            next j
            aadd( ::aData, aclone( aLine ) )
         next i
      endif   
   endif
return self   

Method Bitmap() Class GraphPlus
   return ::hBitMap   

METHOD GraphColors( aColors ) CLASS GraphPlus
   if PCount() == 0
      return ::aColors
   else
      ::aColors := aColors
   endif
   return Self
   
METHOD Legends( aLegends ) CLASS GraphPlus
   if PCount() == 0
      return ::aLegends
   else
      ::aLegends := aLegends
   endif
   return Self
   
METHOD Categories( aCategories ) CLASS GraphPlus
   if PCount() == 0
      return ::aCategories
   else
      ::aCategories := aCategories
   endif
   return Self

METHOD BackColor( aBackColor ) CLASS GraphPlus
   if PCount() == 0
      return ::aBackColor
   else
      ::aBackColor := aBackColor
   endif
   return Self
   
METHOD GraphType( nGraphType ) CLASS GraphPlus
   if PCount() == 0
      return ::nGraphType
   else
      ::nGraphType := nGraphType
   endif
   return Self

METHOD Title( cTitle ) CLASS GraphPlus
   if PCount() == 0
      return ::cTitle
   else
      ::cTitle := cTitle
   endif
   return Self
   
METHOD XTitle( cXTitle ) CLASS GraphPlus
   if PCount() == 0
      return ::cXTitle
   else
      ::cXTitle := cXTitle
   endif
   return Self
      
METHOD YTitle( cYTitle ) CLASS GraphPlus
   if PCount() == 0
      return ::cYTitle
   else
      ::cYTitle := cYTitle
   endif
   return Self
   
METHOD Height( nHeight ) CLASS GraphPlus
   if PCount() == 0
      return ::nImageHeight
   else
      ::nImageHeight := nHeight
   endif
   return Self

METHOD Width( nWidth ) CLASS GraphPlus
   if PCount() == 0
      return ::nImageWidth
   else
      ::nImageWidth := nWidth
   endif
   return Self

METHOD ShowLegends( lShowLegends ) CLASS GraphPlus
   if PCount() == 0
      return ::lShowLegends
   else
      ::lShowLegends := lShowLegends
   endif
   return Self

METHOD ShowLabels( lDataLabels ) CLASS GraphPlus
   if PCount() == 0
      return ::lDataLabels
   else
      ::lDataLabels := lDataLabels
   endif
   return Self

METHOD ShowValues( lShowValues ) CLASS GraphPlus
   if PCount() == 0
      return ::lShowValues
   else
      ::lShowValues := lShowValues
   endif
   return Self

METHOD ShowHGrid( lHGrid ) CLASS GraphPlus
   if PCount() == 0
      return ::lHGrid
   else
      ::lHGrid := lHGrid
   endif
   return Self

METHOD ShowVGrid( lVGrid ) CLASS GraphPlus
   if PCount() == 0
      return ::lVGrid
   else
      ::lVGrid := lVGrid
   endif
   return Self
   
METHOD Picture( cPicture ) CLASS GraphPlus
   if PCount() == 0
      return ::cPicture
   else
      ::cPicture := cPicture
   endif
   return Self
   

Method LegendPos( nLegendPos ) CLASS GraphPlus
   if PCount() == 0
      return ::nLegendPos
   else
      ::nLegendPos := nLegendPos
   endif
   return Self

Method ColorTheme( nColorTheme ) CLASS GraphPlus
   if PCount() == 0
      return ::nColorTheme
   else
      ::nColorTheme := nColorTheme
   endif
   return Self

   
Method Draw CLASS GraphPlus
   local aSize := {}, nLowerUpperRatio := 0
   local nDistance := 0
   local nRow := 0, nPixels := 0
   local nCount := 0
   ::nTop := 10
   ::nLeft := 10
   ::nHeight := ::nImageHeight
   ::nWidth := ::nImageWidth
   ::nBottom := ::nHeight - 10
   ::nRight := ::nWidth - 10
   // init bitmap
   ::hBitmap := BT_BitmapCreateNew ( ::nWidth, ::nHeight, ::aBackColor )
   ::hDC := BT_CreateDC( ::hBitmap, BT_HDC_BITMAP, @::BTStruct )
   
   ::aTextColor := GetNegativeColor ( ::aBackColor )
   
   // draw outside border
   if ::lOutBorder
      BT_DrawRectangle (::hDC, 0, 0, ::nWidth - 1 , ::nHeight - 1, ::aTextColor, 1 )
   endif   
   ::InitColors()    // init colors
   // draw title
   if hmg_len( ::cTitle ) > 0 
      aSize := BT_DrawTextSize (::hDC, ::cTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ], if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ))
      BT_DrawText (::hDC, ::nTop / 2, ::nWidth / 2, ::cTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ], ::aTitleColor, ::aBackColor, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
      ::nTop := ::nTop + aSize[ 2 ]
   else
      ::nTop := ::nTop
   endif
   
   // draw legends
   if ::lShowLegends .and. hmg_len( ::aLegends ) > 0
      ::DrawLegends()
   endif   
   if ::nGraphType <> 4 // not pie
      ::DrawXYTitles()    // draw x/y titles
      // find max value
      // draw graph border
      && BT_DrawRectangle (::hDC, ::nTop, ::nLeft, ::nRight - ::nLeft , ::nBottom - ::nTop, ::aTextColor, 1 )
      
      nMax := 0
      FOR nJ := 1 TO HMG_LEN( ::aData )
         FOR nI :=1 TO HMG_LEN( ::aData[ nJ ] )
            nMax := Max( ::aData[nJ][nI], nMax )
         NEXT nI
      NEXT nJ
      // find min value
      nMin := 0
      FOR nJ := 1 TO HMG_LEN( ::aData )
         FOR nI :=1 TO HMG_LEN( ::aData[nJ])
            nMin := Min( ::aData[nJ][nI], nMin )
         NEXT nI
      NEXT nJ
      if nMax > 0
         ::nUpper := nMax
      else 
         ::nUpper := 0
      endif   
      if nMin < 0
         ::nLower := nMin
      else
         ::nLower := 0   
      endif
      ::nRange := ::nUpper - ::nLower
      if ::nUpper <> 0 .and. ::nLower <> 0
         if abs( ::nLower ) < ::nUpper 
            nLowerUpperRatio := abs( ::nLower ) / ::nRange
            ::nLowerDivision := ::nVDivision * nLowerUpperRatio
            if ::nLowerDivision - int( ::nLowerDivision ) > 0.0
               ::nLowerDivision := int( ::nLowerDivision ) + 1
               ::nUpperDivision := ::nVDivision - ::nLowerDivision
            endif   
            nDivision := ::nUpper / ::nUpperDivision
         else
            nLowerUpperRatio := abs( ::nUpper ) / ::nRange
            ::nUpperDivision := ::nVDivision * nLowerUpperRatio
            if ::nUpperDivision - int( ::nUpperDivision ) > 0.0
               ::nUpperDivision := int( ::nUpperDivision ) + 1
               ::nLowerDivision := ::nVDivision - ::nUpperDivision
            endif   
            nDivision := abs( ::nLower ) / ::nLowerDivision
         endif
      else
         nDivision := ::nRange / ::nVDivision
         if ::nUpper > 0
            ::nUpperDivision := ::nVDivision
            ::nLowerDivision := 0
         else
            ::nLowerDivision := ::nVDivision
            ::nUpperDivision := 0
         endif         
      endif         
      if int( nDivision ) > 0
         cDivision := alltrim( str( int( nDivision ), 14, 0 ) )
         if hmg_len( cDivision ) > 1 // if more than 1 digits round to power of 10s
            nDivision := ( val( left( cDivision, 2 ) ) + 1 ) * 10^ ( hmg_len( cDivision ) - 2 )
         else
            if nDivision > int( nDivision ) 
               nDivision++
            endif
         endif
      else //
         cDivision := transform( nDivision, '9.9999999' )
         nDecimalLength := 8 - hmg_len( alltrim( str( val( right( cDivision, 7 ) ) ) ) )
         nDecimal := val( right( cDivision, 7 ) )
         cDecimal := alltrim( str( nDecimal ) )
         nDivision := (val( alltrim( left( cDecimal, 1 ) )) + 1 ) /(10 ^nDecimalLength)
      endif
      ::nRange := ::nVDivision * nDivision      
      ::nTop := ::nTop + 20
      ::nBottom := ::nBottom - 20
      ::nRight := ::nRight - 5
      if ::lShowCategories 
         nMaxHeight := 0
         for i := 1 to hmg_len( ::aCategories )
            aSize := BT_DrawTextSize (::hDC, ::aCategories[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
            nMaxHeight := max( nMaxHeight, aSize[ 2 ] )
         next i 
         ::nCategoryRow := ::nBottom
         ::nBottom := ::nBottom - nMaxHeight
      else
         ::nBottom := ::nBottom - 20
      endif
      ::nLeft := ::nLeft + 10 
      ::nPixelperPoint := ( ( ::nBottom - ::nTop )  / ::nRange )
      // find zero position
      if ::nLower == 0 // all values are in positive
         if ::lShowCategories .and. hmg_len( ::aCategories ) > 0
            ::lCategoryBelow := .t.
            ::nZeroPos := ::nBottom
         endif   
      elseif ::nUpper == 0 // all values are in negative
         if ::lShowCategories .and. hmg_len( ::aCategories ) > 0
            ::lCategoryBelow := .f.
            ::nZeroPos := ::nTop
         endif   
      else  // some values +ve and some -ve
         ::nZeroPos := ::nTop + ( ::nPixelperPoint * ::nRange/::nVDivision * ::nUpperDivision )
      endif
      asize( ::aScale, 0 )
      nDistance := ::nRange / ::nVDivision
      nRow := ::nTop
      nPixels := ( ::nBottom - ::nTop ) / ::nVDivision
      for i := ::nUpperDivision to 1 step -1
         aadd( ::aScale, { alltrim( transform( nDistance * i, ::cPicture ) ), nRow } )
         nRow := nRow + nPixels
      next i
      aadd( ::aScale, { alltrim( transform( 0, ::cPicture ) ), nRow } )
      nRow := nRow + nPixels
      nCount := 1
      for i := ::nLowerDivision to 1 step -1
         aadd( ::aScale, { '-' + alltrim( transform( nDistance * nCount, ::cPicture ) ), nRow } )
         nCount++
         nRow := nRow + nPixels
      next i
      // vertical scale
      if ::lShowValues
         ::DrawVerticalScale()
      endif
      if ::lHGrid 
         ::DrawHorizontalGrid()
      endif
      if ::lShowCategories
         ::DrawCategories()
      endif
      if ::lVGrid 
         ::DrawVerticalGrid()
      endif
   endif   
   do case
   case ::nGraphType == 1 // Columns
      ::DrawColumnGraph()
   case ::nGraphType == 2 // Line
      ::DrawLineGraph()
   case ::nGraphType == 3 // Points
      ::DrawPointsGraph()
   case ::nGraphType == 4 // Pie
      ::DrawPieGraph()
   case ::nGraphType == 5 // Bar
      ::DrawBarGraph()
   endcase   
   BT_DeleteDC( ::BTstruct )
   
return nil   

Method DrawColumnGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nCategoryWidth := 0
   local nFactor := 0
   local nBarGap := 0
   
   
   if hmg_len( ::aData ) > 0
      if valtype( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      endif
   endif
   if nCategories == 0
      return nil
   endif
   nTotalBars := nCategories * hmg_len( ::aData )
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nTotalBars * 1 // 100%
   nTotalBarRatio := nTotalBarRatio + ( nTotalBarRatio - nCategories ) * ::nBarGapRatio 
   nCategoryRatio := nCategories * ::nGapWidthRatio // n Times of Bar size for each category
   nTotalBarRatio := nTotalBarRatio + nCategoryRatio

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth * ::nGapWidthRatio
   nBarGap      := nBarWidth * ::nBarGapRatio
   
   nCol := ::nLeft + ( nCategoryWidth / 2 )
   nRow := ::nZeroPos
   for i := 1 to hmg_len( ::aData[ 1 ] )
      for j := 1 to hmg_len( ::aData )
         nData := ::aData[ j, i ]
         nBarHeight := abs( nData ) * ::nPixelperPoint
         if nData > 0
            // positive
            BT_DrawFillRectangle ( ::hDC, nRow - nBarHeight, nCol, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow - nBarHeight - 15, nCol + (nBarWidth / 2 ), alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
         else
            // negative
            BT_DrawFillRectangle ( ::hDC, nRow, nCol, nBarWidth, nBarHeight, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + nBarHeight + 3, nCol + (nBarWidth / 2 ), if( nData < 0, '-', '' ) + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
         endif
         nCol := nCol + nBarWidth 
         if j < hmg_len( ::aData )
            nCol := nCol + nBarGap
         endif   
      next j
      nCol := nCol + nCategoryWidth
   next i
return nil

Method DrawVerticalScale Class GraphPlus
   local aSize := {}, i := 0
   local nMaxWidth := 0
   for i := 1 to hmg_len( ::aScale )
      aSize := BT_DrawTextSize (::hDC, ::aScale[ i, 1 ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
      nMaxWidth := max( nMaxWidth, aSize[ 1 ] )
   next i  
   ::nLeft := ::nLeft + nMaxWidth
   for i := 1 to hmg_len( ::aScale )
      BT_DrawText ( ::hDC, ::aScale[ i, 2 ] + 2, ::nLeft, ::aScale[ i, 1 ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ),  BT_TEXT_BASELINE + BT_TEXT_RIGHT)
   next i 
   ::nLeft := ::nLeft + 5
//   BT_DrawLine (::hDC, ::nTop, nCol + 5, ::nBottom, nCol + 5, ::aTextColor, ::nLineWidth)
return nil

Method DrawHorizontalGrid Class GraphPlus
   local i := 0
   for i := 1 to hmg_len( ::aScale )
      BT_DrawLine (::hDC, ::aScale[ i, 2 ], ::nLeft, ::aScale[ i, 2 ], ::nRight, ::aGridLineColor, ::nLineWidth)
   next i
return nil

Method DrawVerticalGrid Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nBarWidth := 0
   local nCategoryWidth := 0
   nCategories := hmg_len( ::aData[ 1 ] )
   if nCategories == 0
      return nil
   endif
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nCategories // equal space between categories
   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth 
   BT_DrawLine (::hDC, ::nTop, ::nLeft, ::nBottom, ::nLeft, ::aGridLineColor, ::nLineWidth)
   nCol := ::nLeft
   for i := 1 to nCategories
      nCol := nCol + nCategoryWidth
      BT_DrawLine (::hDC, ::nTop, nCol, ::nBottom, nCol, ::aGridLineColor, ::nLineWidth)
   next i
return nil


Method DrawLineGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0, nFirstCol := 0, nPrevRow := 0, nPrevCol := 0
   local nTotalBars := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nCategoryWidth := 0
   local nFactor := 0
   local nBarGap := 0
   
   if hmg_len( ::aData ) > 0
      if valtype( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      endif
   endif
   if nCategories == 0
      return nil
   endif
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nCategories // equal space between categories

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth 
   nFirstCol := ::nLeft + ( nCategoryWidth / 2 )   
   nCol := ::nLeft + ( nCategoryWidth / 2 )
   nRow := ::nZeroPos
   for j := 1 to hmg_len( ::aData )
      nCol := nFirstCol
      for i := 1 to hmg_len( ::aData[ 1 ] )
         nData := ::aData[ j, i ]
         nBarHeight := abs( nData ) * ::nPixelperPoint
         if nData > 0
            // positive
            BT_DrawFillEllipse ( ::hDC, nRow - nBarHeight - ::nPointSize/2, nCol - ::nPointSize / 2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1)
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow - nBarHeight - 15 - ::nPointSize / 2 , nCol, alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
            if i <> 1 // first point
               BT_DrawLine (::hDC, nPrevRow, nPrevCol, nRow - nBarHeight, nCol, ::aColors[ j ], ::nLineWidth)              
            endif
            nPrevRow := nRow - nBarHeight
            nPrevCol := nCol
         else
            // negative
            BT_DrawFillEllipse ( ::hDC, nRow + nBarHeight - ::nPointSize/2, nCol - ::nPointSize/2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + nBarHeight + 3 + ::nPointSize / 2 , nCol, if( nData < 0, '-', '' ) + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
            if i <> 1
               BT_DrawLine (::hDC, nPrevRow, nPrevCol, nRow + nBarHeight, nCol, ::aColors[ j ], ::nLineWidth)              
            endif
            nPrevRow := nRow + nBarHeight
            nPrevCol := nCol
         endif
         nCol := nCol + nCategoryWidth
      next i
   next j

return nil

Method DrawPointsGraph Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nCategoryWidth := 0
   local nFactor := 0
   local nBarGap := 0
   
   if hmg_len( ::aData ) > 0
      if valtype( ::aData[ 1 ] ) == 'A'
         nCategories := hmg_len( ::aData[ 1 ] )
      endif
   endif
   if nCategories == 0
      return nil
   endif
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nCategories // equal space between categories

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth 
   
   nCol := ::nLeft + ( nCategoryWidth / 2 )
   nRow := ::nZeroPos
   for i := 1 to hmg_len( ::aData[ 1 ] )
      for j := 1 to hmg_len( ::aData )
         nData := ::aData[ j, i ]
         nBarHeight := abs( nData ) * ::nPixelperPoint
         if nData > 0
            // positive
            BT_DrawFillEllipse ( ::hDC, nRow - nBarHeight - ::nPointSize / 2, nCol - ::nPointSize / 2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1)
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow - nBarHeight - 15 - ::nPointSize / 2 , nCol, alltrim( transform( nData, ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
         else
            // negative
            BT_DrawFillEllipse ( ::hDC, nRow + nBarHeight - ::nPointSize / 2, nCol - ::nPointSize / 2, ::nPointSize, ::nPointSize, ::aColors[ j ], ::aColors[ j ], 1 )
            if ::lDataLabels
               BT_DrawText ( ::hDC, nRow + nBarHeight + 3 + ::nPointSize / 2 , nCol, if( nData < 0, '-', '' ) + alltrim( transform( abs( nData ), ::cPicture ) ), ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER )
            endif
         endif
      next j
      nCol := nCol + nCategoryWidth
   next i
return nil

Method DrawPieGraph Class GraphPlus
   LOCAL fromrow := 0
   LOCAL fromcol := 0
   LOCAL torow := 0
   LOCAL tocol := 0
   LOCAL topleftrow
   LOCAL topleftcol
   LOCAL toprightrow
   LOCAL toprightcol
   LOCAL bottomrightrow
   LOCAL bottomrightcol
   LOCAL bottomleftrow
   LOCAL bottomleftcol
   LOCAL middletopcol
   LOCAL middleleftrow
   LOCAL middleleftcol
   LOCAL middlebottomcol
   LOCAL middlerightrow
   LOCAL middlerightcol
   LOCAL fromradialrow
   LOCAL fromradialcol
   LOCAL toradialrow
   LOCAL toradialcol
   LOCAL degrees := {}
   LOCAL cumulative := {}
   LOCAL j,i,sum := 0
   LOCAL shadowcolor
   LOCAL previos_cumulative
   local aSerieValues := {}
   local nChartSize := 0
   local aLabelCumulative := {}
   
   for i := 1 to hmg_len( ::aData[ 1 ] )
      aadd( aSerieValues, ::aData[ 1, i ] )
   next i   
   ::nLeft += 25
   ::nRight -= 25
   ::nBottom -= 25
   ::nTop += 25
   nChartSize :=  min( ::nRight - ::nLeft, ::nBottom - ::nTop ) 

   fromcol := ::nLeft + ( ( ::nRight - ::nLeft ) - nChartSize ) / 2
   tocol := ::nRight - ( ( ::nRight - ::nLeft ) - nChartSize ) / 2
   torow := ::nBottom - ( ( ::nBottom - ::nTop ) - nChartSize ) / 2
   fromrow := ::nTop + ( ( ::nBottom - ::nTop ) - nChartSize ) / 2
   
   topleftrow := fromrow
   topleftcol := fromcol
   toprightrow := fromrow
   toprightcol := tocol
   bottomrightrow := torow
   bottomrightcol := tocol
   bottomleftrow := torow
   bottomleftcol := fromcol
   middletopcol := fromcol + int(tocol - fromcol) / 2
   middleleftrow := fromrow + int(torow - fromrow) / 2
   middleleftcol := fromcol
   middlebottomcol := fromcol + int(tocol - fromcol) / 2
   middlerightrow := fromrow + int(torow - fromrow) / 2
   middlerightcol := tocol

   torow := torow + 1
   tocol := tocol + 1
   for i := 1 to HMG_LEN(aSerieValues)
      sum := sum + aSerieValues[i]
   next i
   for i := 1 to HMG_LEN(aSerieValues)
      aadd(degrees,Round(aSerieValues[i]/sum * 360,0))
   next i
   sum := 0
   for i := 1 to HMG_LEN(degrees)
      sum := sum + degrees[i]
   next i
   if sum <> 360
      degrees[HMG_LEN(degrees)] := degrees[HMG_LEN(degrees)] + (360 - sum)
   endif

   sum := 0
   lSum := 0
   for i := 1 to HMG_LEN(degrees)
      sum := sum + degrees[i]
      aadd(cumulative,sum)
      lSum := lSum + ( degrees[i] / 2 )
      aadd( aLabelCumulative, lSum )
      lSum := lSum + ( degrees[i] / 2 )
   next i

   previos_cumulative := -1

   fromradialrow := middlerightrow
   fromradialcol := middlerightcol
   for i := 1 to HMG_LEN(cumulative)

      if cumulative[i] == previos_cumulative
         loop
      endif

      previos_cumulative := cumulative[i]
      do case
      case cumulative[i] <= 45
         toradialcol := middlerightcol
         toradialrow := middlerightrow - Round(cumulative[i] / 45 * (middlerightrow - toprightrow),0)
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      case cumulative[i] <= 90 .and. cumulative[i] > 45
         toradialrow := toprightrow
         toradialcol := toprightcol - Round((cumulative[i] - 45) / 45 * (toprightcol - middletopcol),0)
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      case cumulative[i] <= 135 .and. cumulative[i] > 90
         toradialrow := topleftrow
         toradialcol := middletopcol - Round((cumulative[i] - 90) / 45 * (middletopcol - topleftcol),0)
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      case cumulative[i] <= 180 .and. cumulative[i] > 135
         toradialcol := topleftcol
         toradialrow := topleftrow + Round((cumulative[i] - 135) / 45 * (middleleftrow - topleftrow),0)
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      case cumulative[i] <= 225 .and. cumulative[i] > 180
         toradialcol := topleftcol
         toradialrow := middleleftrow + Round((cumulative[i] - 180) / 45 * (bottomleftrow - middleleftrow),0)
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      case cumulative[i] <= 270 .and. cumulative[i] > 225
         toradialrow := bottomleftrow
         toradialcol := bottomleftcol + Round((cumulative[i] - 225) / 45 * (middlebottomcol - bottomleftcol),0)
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      case cumulative[i] <= 315 .and. cumulative[i] > 270
         toradialrow := bottomleftrow
         toradialcol := middlebottomcol + Round((cumulative[i] - 270) / 45 * (bottomrightcol - middlebottomcol),0)
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      case cumulative[i] <= 360 .and. cumulative[i] > 315
         toradialcol := bottomrightcol
         toradialrow := bottomrightrow - Round((cumulative[i] - 315) / 45 * (bottomrightrow - middlerightrow),0)
         BT_DrawPie ( ::hDC, fromrow, fromcol, torow, tocol, fromradialrow, fromradialcol, toradialrow, toradialcol, ::aColors[ i ], 1, ::aColors[ i ] )
         fromradialrow := toradialrow
         fromradialcol := toradialcol
      endcase
   next i
   if ::lShowValues
      nChartSize :=  nChartSize / 1.5 

      fromcol := ::nLeft + ( ( ::nRight - ::nLeft ) - nChartSize ) / 2
      tocol := ::nRight - ( ( ::nRight - ::nLeft ) - nChartSize ) / 2
      torow := ::nBottom - ( ( ::nBottom - ::nTop ) - nChartSize ) / 2
      fromrow := ::nTop + ( ( ::nBottom - ::nTop ) - nChartSize ) / 2
      
      topleftrow := fromrow
      topleftcol := fromcol
      toprightrow := fromrow
      toprightcol := tocol
      bottomrightrow := torow
      bottomrightcol := tocol
      bottomleftrow := torow
      bottomleftcol := fromcol
      middletopcol := fromcol + int(tocol - fromcol) / 2
      middleleftrow := fromrow + int(torow - fromrow) / 2
      middleleftcol := fromcol
      middlebottomcol := fromcol + int(tocol - fromcol) / 2
      middlerightrow := fromrow + int(torow - fromrow) / 2
      middlerightcol := tocol

      torow := torow + 1
      tocol := tocol + 1
   
   
      previos_cumulative := -1


      fromradialrow := middlerightrow
      fromradialcol := middlerightcol
      for i := 1 to HMG_LEN(cumulative)

         if aLabelCumulative[i] == previos_cumulative
            loop
         endif
         cValue := alltrim(transform( aSerieValues[i], ::cPicture)) + '(' + alltrim( transform( degrees[ i ] / 3.60, '999.99' ) ) + '%)'  
         previos_cumulative := aLabelCumulative[i]
         do case
         case aLabelCumulative[i] <= 45
            toradialcol := middlerightcol
            toradialrow := middlerightrow - Round(aLabelCumulative[i] / 45 * (middlerightrow - toprightrow),0)
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
            fromradialrow := toradialrow
            fromradialcol := toradialcol
         case aLabelCumulative[i] <= 90 .and. aLabelCumulative[i] > 45
            toradialrow := toprightrow
            toradialcol := toprightcol - Round((aLabelCumulative[i] - 45) / 45 * (toprightcol - middletopcol),0)
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
            fromradialrow := toradialrow
            fromradialcol := toradialcol
         case aLabelCumulative[i] <= 135 .and. aLabelCumulative[i] > 90
            toradialrow := topleftrow
            toradialcol := middletopcol - Round((aLabelCumulative[i] - 90) / 45 * (middletopcol - topleftcol),0)
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
            fromradialrow := toradialrow
            fromradialcol := toradialcol
         case aLabelCumulative[i] <= 180 .and. aLabelCumulative[i] > 135
            toradialcol := topleftcol
            toradialrow := topleftrow + Round((aLabelCumulative[i] - 135) / 45 * (middleleftrow - topleftrow),0)
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
            fromradialrow := toradialrow
            fromradialcol := toradialcol
         case aLabelCumulative[i] <= 225 .and. aLabelCumulative[i] > 180
            toradialcol := topleftcol
            toradialrow := middleleftrow + Round((aLabelCumulative[i] - 180) / 45 * (bottomleftrow - middleleftrow),0)
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
            fromradialrow := toradialrow
            fromradialcol := toradialcol
         case aLabelCumulative[i] <= 270 .and. aLabelCumulative[i] > 225
            toradialrow := bottomleftrow
            toradialcol := bottomleftcol + Round((aLabelCumulative[i] - 225) / 45 * (middlebottomcol - bottomleftcol),0)
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
            fromradialrow := toradialrow
            fromradialcol := toradialcol
         case aLabelCumulative[i] <= 315 .and. aLabelCumulative[i] > 270
            toradialrow := bottomleftrow
            toradialcol := middlebottomcol + Round((aLabelCumulative[i] - 270) / 45 * (bottomrightcol - middlebottomcol),0)
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
            fromradialrow := toradialrow
            fromradialcol := toradialcol
         case aLabelCumulative[i] <= 360 .and. aLabelCumulative[i] > 315
            toradialcol := bottomrightcol
            toradialrow := bottomrightrow - Round((aLabelCumulative[i] - 315) / 45 * (bottomrightrow - middlerightrow),0)
            BT_DrawText ( ::hDC, toradialrow, toradialcol, cValue, ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aColors[ i ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
            fromradialrow := toradialrow
            fromradialcol := toradialcol
         endcase
      next i
   endif
   
return nil

Method DrawBarGraph Class GraphPlus

return nil

Method Save( cFileName ) Class GraphPlus
   if pcount() == 0
      cFileName := GetFile ()
   endif
   cFileName := alltrim( cFileName )
   if hmg_len( cFileName ) > 0   
      if file( cFileName ) 
         if .not. msgyesno( 'File already Exists! Do you want to overwrite?' )
            return nil
         endif
      endif
      BT_BitmapSaveFile( ::hBitMap, cFileName, BT_FILEFORMAT_PNG )
   endif      
return nil


Method DrawLegends Class GraphPlus
   local nMaxWidth := 0, i := 0, nMaxHeight := 0
   local nLegends := hmg_len( ::aLegends )
   local aSize := {}
   local nColumns := 0, nRows := 0
   local nRow := 0
   local nCol := 0
   for i := 1 to nLegends
      aSize := BT_DrawTextSize (::hDC, ::aLegends[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
      nMaxWidth := max( nMaxWidth, aSize[ 1 ] )
      nMaxHeight := max( nMaxHeight, aSize[ 2 ] )
   next i   
   nMaxWidth := ( nMaxWidth + ::nLegendBoxSize + 10 ) // per legend
   nMaxHeight := nMaxHeight + 10                      // per legend
   // find columns/rows
   if ::nLegendPos == 1 // bottom
      nColumns := int( ( ::nWidth -  20 ) / nMaxWidth )
      if nColumns == 0 // legend length is not fitting!
         nColumns := 1
      endif
      nRows := int( nLegends / nColumns ) + ( if( mod( nLegends, nColumns ) > 0, 1, 0 ) )
   else // right
      nColumns := 1
      nRows := nLegends
      if nLegends * nMaxHeight > ( ::nBottom - ::nTop )
         // all the legends can not fit!
         nRows := int( ( ::nBottom - ::nTop ) / nMaxHeight )
      endif
   endif   
   // draw legends
   if ::nLegendPos == 1 // bottom
      ::nBottom := ::nBottom - ( nRows * nMaxHeight )
      nRow := ::nBottom + 5
      i := 1
      j := 1
      nCol := 5
      do while i <= nLegends
         if j > nColumns
            j := 1
            nRow := nRow + nMaxHeight
            nCol := 5
         endif
         BT_DrawFillRectangle ( ::hDC, nRow, nCol, ::nLegendBoxSize, max( nMaxHeight - 10, ::nLegendBoxSize ), ::aColors[ i ], ::aColors[ i ], 1 )
         BT_DrawText ( ::hDC, nRow, nCol + ::nLegendBoxSize + 5, ::aLegends[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
         nCol := nCol + nMaxWidth
         j++
         i++
      enddo
   else  // right
      ::nRight := ::nRight - nMaxWidth
      nRow := ::nTop + 5
      nCol := ::nRight + 5
      for i := 1 to nRows
         BT_DrawFillRectangle ( ::hDC, nRow, nCol, ::nLegendBoxSize, max( nMaxHeight - 10, ::nLegendBoxSize ), ::aColors[ i ], ::aColors[ i ], 1 )
         BT_DrawText ( ::hDC, nRow, nCol + ::nLegendBoxSize + 5, ::aLegends[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ))
         nRow := nRow + nMaxHeight         
      next i   
   endif   
return nil

Method DrawCategories Class GraphPlus
   local nCategories := 0
   local nRow := 0, nCol := 0, nData := 0
   local nTotalBars := 0
   local nAvailableWidth := 0
   local nTotalBarRatio := 0
   local nCategoryRatio := 0
   local nTotalBarSize := 0

   local nBarWidth := 0
   local nCategoryWidth := 0
   local nFactor := 0
   local nBarGap := 0
   nCategories := hmg_len( ::aCategories )
   if nCategories == 0
      return nil
   endif
   nAvailableWidth := ::nRight - ::nLeft
   nTotalBarRatio := nCategories // equal space between categories

   nBarWidth := nAvailableWidth / nTotalBarRatio // 100% width
   nCategoryWidth := nBarWidth 
   
   nCol := ::nLeft + ( nCategoryWidth / 2 )
   nRow := ::nZeroPos
   for i := 1 to hmg_len( ::aCategories )
      BT_DrawText ( ::hDC, ::nCategoryRow, nCol, ::aCategories[ i ], ::aLegendFont[ 1 ], ::aLegendFont[ 2 ], ::aTextColor, ::aBackColor, if( ::aLegendFont[ 3 ], BT_TEXT_BOLD, nil ),  BT_TEXT_CENTER)
      nCol := nCol + nCategoryWidth
   next i
return nil

Method DrawXYTitles Class GraphPlus
   local aSize := {}
   if hmg_len( ::cXTitle ) > 0
      aSize := BT_DrawTextSize (::hDC, ::cXTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2 , if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ))
      BT_DrawText (::hDC, ::nBottom - aSize[ 2 ], (::nRight - ::nLeft) / 2, ::cXTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2, ::aTitleColor, ::aBackColor, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER)
      ::nBottom := ::nBottom - aSize[ 2 ]
   endif
   if hmg_len( ::cYTitle ) > 0
      aSize := BT_DrawTextSize (::hDC, ::cYTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2 , if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ))
      BT_DrawText (::hDC, (::nBottom - ::nTop) / 2, ::nLeft, ::cYTitle, ::aTitleFont[ 1 ], ::aTitleFont[ 2 ] - 2, ::aTitleColor, ::aBackColor, if( ::aTitleFont[ 3 ], BT_TEXT_BOLD, nil ), BT_TEXT_CENTER, BT_TEXT_VERTICAL_ASCENDANT)
      ::nLeft := ::nLeft + aSize[ 2 ]
   endif      
return nil



Method InitColors Class GraphPlus
   local i := 0, j := 0, nLegends := 0, nFactor := 0, nCurFactor := 1
   local aHSL := {}
   local aColorTheme := {}
   nLegends := hmg_len( ::aLegends )
   if hmg_len( ::aColors ) = 0 .or. hmg_len( ::aColors ) < nLegends
      asize( ::aColors, 0 )
      aColorTheme := ::aColorThemes[ ::nColorTheme ]
      if nLegends <= hmg_len( aColorTheme )
         for i := 1 to nLegends
            aadd( ::aColors, aColorTheme[ i ] )
         next i
      else
         for i := 1 to hmg_len( aColorTheme )
            aadd( ::aColors, aColorTheme[ i ] )         
         next i
         i := hmg_len( aColorTheme ) + 1
         j := 1
         if ::lLighter
            nFactor := 1.03
         else
            nFactor := 0.97
         endif            
         do while i <= hmg_len( ::aLegends )
            if j > hmg_len( aColorTheme )
               j := 1
               nCurFactor := nCurFactor * nFactor
            endif   
            aHSL := GP_RGB2HSL( aColorTheme[ j ] )
            aHSL[ 3 ] := aHSL[ 3 ] * nCurFactor
            aadd( ::aColors, GP_HSL2RGB( aHSL ) )
            j++
            i++
         enddo
      endif
   endif
return nil

Method SwitchRowsToColumns
   local aTemp := {}, aLine := {}
   local cTemp := ''
   local i := 0
   local j := 0
   // transpose data
   for i := 1 to hmg_len( ::aData[ 1 ] )
      asize( aLine, 0 )
      for j := 1 to hmg_len( ::aData )
         aadd( aLine, ::aData[ j, i ] )
      next j
      aadd( aTemp, aclone( aLine ) )
   next i
   asize( ::aData, 0 )   
   ::aData := aclone( aTemp )
   // switch legends and categories
   aTemp := aclone( ::aCategories )
   ::aCategories := aclone( ::aLegends )
   ::aLegends := aclone( aTemp )
   // switch xy titles
   cTemp := ::cXTitle
   ::cXTitle := ::cYTitle
   ::cYTitle := cTemp
return nil   




function findboundary( nNum )
   LOCAL nE, nMax, nMan, nVal, nOffset

   nE:=10
   nVal:=0
   nNum:=Abs(nNum)

   DO WHILE .T.

      nMax := 10**nE

      IF Int(nNum/nMax)>0

         nMan:=(nNum/nMax)-Int(nNum/nMax)
         nOffset:=1
         nOffset:=IF(nMan<=.75,.75,nOffset)
         nOffset:=IF(nMan<=.50,.50,nOffset)
         nOffset:=IF(nMan<=.25,.25,nOffset)
         nOffset:=IF(nMan<=.00,.00,nOffset)
         nVal := (Int(nNum/nMax)+nOffset)*nMax
         EXIT

      ENDIF

      nE--

   ENDDO

RETURN (nVal)
   

function GetNegativeColor( aRGB )
   local aNewRGB := {}
   aNewRGB := { 255 - aRGB[ 1 ], 255 - aRGB[ 2 ], 255 - aRGB[ 3 ] }
return aNewRGB
   
FUNCTION GP_RGB2HSL( aRGB )

   LOCAL nMax, nMin, nDelta
   LOCAL nH, nS, nL, nR, nG, nB

   nR := aRGB[1] / 255
   nG := aRGB[2] / 255
   nB := aRGB[3] / 255
   nMax := Max( nR, Max( nG, nB ) )
   nMin := Min( nR, Min( nG, nB ) )
   nDelta := nMax - nMin

   IF nDelta == 0
      nH := 0
   ELSEIF ( nMax == nR )
      nH := (( nG - nB) / nDelta) % 6
   ELSEIF ( nMax == nG )
      nH := (nB - nR) / nDelta + 2
   ELSE
      nH := (nR - nG) / nDelta + 4
   ENDIF
   nH := Int( nH * 60 )
   IF ( nH < 0 )
      nH += 360
   ENDIF
   nL := ( nMax + nMin ) / 2
   nS := iif( nDelta == 0, 0, nDelta / ( 1 - Abs( 2 * nL - 1 )) )
   nS := nS * 100
   nL := nL * 100
   RETURN { nH, nS, nL }

FUNCTION GP_HSL2RGB( aHSL )

   LOCAL nR := 0, nG := 0, nB := 0, nH, nS, nL, c, x, m
   nH := aHSL[1]
   nS := aHSL[2] / 100
   nL := aHSL[3] / 100
   c := ( 1 - Abs( 2 * nL - 1 ) ) * nS
   x := c * ( 1 - Abs( ( nH / 60 ) % 2 - 1 ) )
   m := nL - c / 2
   IF (0 <= nH .And. nH < 60)
      nR := c
      nG := x
      nB := 0
   ELSEIF (60 <= nH .And. nH < 120)
      nR := x
      nG := c
      nB := 0
   ELSEIF (120 <= nH .And. nH < 180)
      nR := 0
      nG := c
      nB := x
   ELSEIF (180 <= nH .And. nH < 240)
      nR := 0
      nG := x
      nB := c
   ELSEIF (240 <= nH .And. nH < 300)
      nR := x
      nG := 0
      nB := c
   ELSEIF (300 <= nH .And. nH < 360)
      nR := c
      nG := 0
      nB := x
   ENDIF
   nR := Int( ( nR + m ) * 255 )
   nG := Int( ( nG + m ) * 255 )
   nB := Int( ( nB + m ) * 255 )
   RETURN { nR, nG, nB }   
   
function degreetoradian( nDegree )
   return ( nDegree * 0.0174532925 )

function radiantodegree( nRadian )
   return ( nRadian / 0.0174532925 )  

   
   